package lc.esp.eo;

import com.mongodb.BasicDBObject;
import com.mongodb.MongoGridFSException;
import com.mongodb.TransactionOptions;
import com.mongodb.WriteConcern;
import com.mongodb.client.ClientSession;
import com.mongodb.client.MongoClient;
import com.mongodb.client.MongoCollection;
import com.mongodb.client.MongoDatabase;
import com.mongodb.client.gridfs.GridFSBucket;
import com.mongodb.client.gridfs.GridFSBuckets;
import com.mongodb.client.gridfs.model.GridFSUploadOptions;
import com.mongodb.client.result.DeleteResult;
import lc.esp.eo.sdk.EO;
import lc.esp.eo.sdk.EOId;
import lc.esp.eo.sdk.schema.AssociationSchema;
import lc.esp.eo.sdk.schema.Model;
import lc.mecha.log.MechaLogger;
import lc.mecha.log.MechaLoggerFactory;
import org.bson.Document;
import org.bson.types.ObjectId;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Serializable;
import java.util.HashSet;
import java.util.Map;

/**
 * This class implements a transaction with the back-end. This includes both functionality for the objects themselves
 * and EOFS.
 *
 * @author Alex Leigh
 * @since mk20 (TOKYO PROTOCOL)
 */
public class EOTransaction implements AutoCloseable {
    private static final MechaLogger logger = MechaLoggerFactory.getLogger(EOTransaction.class);
    private final ClientSession session;
    private final MongoClient client;
    private final String databaseName;
    private final MongoDatabase db;
    private final static String KEY_ID = "_id";
    private final static String KEY_DATA = "data";
    private final static String KEY_TYPE = "type";
    private final static String KEY_DSTOBJ = "dstObj";
    private final static String KEY_SRCOBJ = "srcObj";
    private final static String KEY_CONTENT_TYPE = "content_type";
    private final static String COLLECTION_OBJ = "eo_obj";
    private final static String COLLECTION_ASSOC = "eo_assoc";
    private final static String COLLECTION_CHUNKS = "chunks";
    private final static String COLLECTION_FILES = "files";
    private final static String KEY_FS_FILENAME = "name";
    private final static String TYPE_FS_LINK = "link";

    private final static String TYPE_INODE = "inode";

    private HashSet<ObjectId> pendingFiles;

    private final Model model;

    public EOTransaction(MongoClient client, String databaseName, Model model) {
        this.client = client;
        this.model = model;
        this.session = this.client.startSession();
        this.session.startTransaction(TransactionOptions.builder().writeConcern(WriteConcern.ACKNOWLEDGED).build());
        this.databaseName = "eo_" + databaseName;
        this.db = client.getDatabase(this.databaseName);
    }

    /**
     * Initialize a new database for use with EO. This method runs outside of the transaction.
     * <p>
     * Note that there is a propagation delay from the time this call is made to when it is consistent across the
     * cluster. If the caller immediately tries to perform a data operation on the collections it may result in
     * errors.
     */
    public void initialize() {
        shardByKey(COLLECTION_OBJ, KEY_ID);
        shardByKey(COLLECTION_ASSOC, KEY_SRCOBJ);
        shardByKey(COLLECTION_FILES, KEY_ID);
        shardByKey(COLLECTION_CHUNKS, KEY_ID);
    }

    /**
     * Shard the given collection to the given key using a hashed strategy. This call is asynchronous and may return
     * before all of the shards have been updated.
     */
    private void shardByKey(String collectionName, String keyName) {
        String path = databaseName + "." + collectionName;
        MongoDatabase adminDb = client.getDatabase("admin");
        Document cmd = new Document("shardCollection", path).append("key", new Document(keyName, "hashed"));
        Document response = adminDb.runCommand(cmd);
        logger.info("shardBykey: {}", response);
        // TODO: Wait for `db.collection.getShardDistribution()`?
    }

    /**
     * Create an object of the given type and return the newly created id.
     */
    public EOId objCreate(String type) {
        return objCreate(type, null);
    }

    /**
     * Create a new object of the given type, with the given user-defined data.
     */
    public EOId objCreate(String type, Map<String, Serializable> values) {
        // FIXME: Returning ObjectId here is a leaky abstraction
        assert (type != null);
        MongoCollection<Document> col = db.getCollection(COLLECTION_OBJ);
        Document d = new Document();
        if (values != null) d.append(KEY_DATA, values);
        d.append(KEY_TYPE, type);
        col.insertOne(session, d);
        ObjectId id = d.getObjectId(KEY_ID);
        return id == null ? null : new EOBsonId(id);
    }

    /**
     * Create an association between the source and destination objects of the given type.
     */
    public String assocCreate(EOId srcObjectId, EOId dstObjectId, String type) {
        return assocCreate(srcObjectId, dstObjectId, type, null);
    }

    /**
     * Create a one-way association between two objects. The association may optionally contain a user-defined
     * data.
     */
    public String assocCreate(EOId srcObjectId, EOId dstObjectId, String type, Map<String, Serializable> values) {
        assert (srcObjectId != null);
        assert (dstObjectId != null);
        assert (type != null);

        MongoCollection<Document> col = db.getCollection(COLLECTION_ASSOC);
        Document d = new Document();
        if (values != null) d.append(KEY_DATA, values);
        d.append(KEY_TYPE, type);
        d.append(KEY_SRCOBJ, ((EOBsonId) srcObjectId).getObjectId());
        d.append(KEY_DSTOBJ, ((EOBsonId) dstObjectId).getObjectId());
        col.insertOne(session, d);
        ObjectId id = d.getObjectId(KEY_ID);

        if (model != null) {
            AssociationSchema as = model.getAssocSchema(type);
            if (as.getInverseType() != null) {
                d = new Document();
                if (values != null) d.append(KEY_DATA, values);
                d.append(KEY_TYPE, as.getInverseType());
                d.append(KEY_SRCOBJ, ((EOBsonId) dstObjectId).getObjectId());
                d.append(KEY_DSTOBJ, ((EOBsonId) srcObjectId).getObjectId());
                col.insertOne(session, d);
            }
        }

        return id == null ? null : id.toString();
    }

    /**
     * Return a count of the associations originating from the given source object of the given type.
     */
    public long assocCount(EOId srcObjectId, String type) {
        MongoCollection<Document> col = db.getCollection(COLLECTION_ASSOC);
        return col.countDocuments(session, new Document()
                .append(KEY_SRCOBJ, ((EOBsonId) srcObjectId).getObjectId())
                .append(KEY_TYPE, type));
    }

    /**
     * Delete the given association. If a model is defined in the transaction and an inverse association is defined,
     * then the inverse will also be deleted. The method returns the count of the associations which were deleted.
     */
    public long assocDelete(EOId srcObjectId, EOId dstObjectId, String type) {
        MongoCollection<Document> col = db.getCollection(COLLECTION_ASSOC);

        long cnt = 0;

        // deleteOne() under the premise that they are unique
        DeleteResult result = col.deleteOne(session, new Document()
                .append(KEY_TYPE, type)
                .append(KEY_SRCOBJ, ((EOBsonId) srcObjectId).getObjectId())
                .append(KEY_DSTOBJ, ((EOBsonId) dstObjectId).getObjectId()));

        cnt += result.getDeletedCount();

        if (model != null) {
            AssociationSchema as = model.getAssocSchema(type);
            if (as != null) {
                result = col.deleteOne(session, new Document()
                        .append(KEY_TYPE, as.getInverseType())
                        .append(KEY_SRCOBJ, ((EOBsonId) dstObjectId).getObjectId())
                        .append(KEY_DSTOBJ, ((EOBsonId) srcObjectId).getObjectId()));
                cnt += result.getDeletedCount();
            }
        }

        return cnt;
    }

    /**
     * Read a given object and return it to the caller in the form of an {@link EO}. Reading the object will not
     * query the underlying associations, so only the id, type, and any user-defined data is returned.
     */
    public EO objRead(EOId id) {
        MongoCollection<Document> col = db.getCollection(COLLECTION_OBJ);
        BasicDBObject query = new BasicDBObject();
        query.put(KEY_ID, ((EOBsonId) id).getObjectId());
        Document doc = col.find(session, query).first();

        if (doc == null) return null;

        EO eo = new EO(doc.getString(KEY_TYPE), new EOBsonId(doc.getObjectId(KEY_ID)));
        Document dataDoc = (Document) doc.get(KEY_DATA);
        for (String key : dataDoc.keySet()) {
            eo.setValue(key, dataDoc.get(key));
        }

        return eo;
    }

    /**
     * Delete the given object.
     */
    public void objDelete(EOId id) {
        // TODO: Implement association delete
        MongoCollection<Document> col = db.getCollection(COLLECTION_OBJ);
        col.deleteOne(session, new Document().append("_id", ((EOBsonId) id).getObjectId()));
    }

    /**
     * Put a file into the database. Note that due to lack of support for multi-document transactions with GridFS,
     * the actual storage of the file is not transactional. If this transaction is rolled back an effort will be made
     * to delete the file; however the result of this operation is not guaranteed. If the file fails to delete for
     * whatever reason, the EOFS directory objects (created on the transaction) will be correct, but the file will
     * actually exist in GridFs, in an unknown condition; either fully or partially intact.
     */
    public String fsFileCreate(String filename, InputStream file, String contentType) {
        // TODO: Implement directory structure
        if (filename.contains("/"))
            throw new IllegalArgumentException("Filename provided may not contain file separator.");

        GridFSUploadOptions options = new GridFSUploadOptions()
                .chunkSizeBytes(1048576)
                .metadata(new Document(KEY_CONTENT_TYPE, contentType));
        GridFSBucket gridFSBucket = GridFSBuckets.create(db);
        ObjectId fileId = gridFSBucket.uploadFromStream(filename, file, options);
        synchronized (this) {
            if (pendingFiles == null) pendingFiles = new HashSet<>();
            pendingFiles.add(fileId);
        }
        return fileId.toString();
    }

    public void fsFileRead(OutputStream os, String filename) throws IOException {
        // TODO: Implement directory structure
        // TODO: Support getting the file-type/size for the file
        GridFSBucket gridFSBucket = GridFSBuckets.create(db);
        gridFSBucket.downloadToStream(session, filename, os);
        os.flush();
    }

    /**
     * Create a new directory and link it to a parent.
     */
    public void fsDirCreate(EOId parentId, String name) {
        EOId newId = objCreate(TYPE_INODE);
        assocCreate(parentId, newId, TYPE_FS_LINK, Map.of(KEY_FS_FILENAME, name));
        assocCreate(newId, newId, TYPE_FS_LINK, Map.of(KEY_FS_FILENAME, "."));
        assocCreate(newId, parentId, TYPE_FS_LINK, Map.of(KEY_FS_FILENAME, ".."));
    }

    /**
     * Commit the current transaction, and begin a new one.
     */
    public void commit() {
        session.commitTransaction();
        pendingFiles = null;
        session.startTransaction();
    }

    /**
     * Roll back the current transaction and start a new one. If this transaction has pending files (files that
     * were created with <code>fileCreate()</code> but have not yet been comitted, then this method will make an
     * effort to delete those files from GridFS, which is not guaranteed. References to the file which were created
     * in the directory objects will be reliably rolled back.
     */
    public void rollback() {
        session.abortTransaction();
        if (pendingFiles != null) {
            GridFSBucket gridFSBucket = GridFSBuckets.create(db);
            for (ObjectId pending : pendingFiles) {
                try {
                    gridFSBucket.delete(pending);
                } catch (MongoGridFSException e) {
                    logger.warn("Failed to delete file created adjacent to pending transaction: {}", e);
                }
            }
        }
        pendingFiles = null;
        session.startTransaction();
    }

    /**
     * Close the given transaction and free any associated resources. If the transaction had pending changes, it will
     * be rolled back.
     */
    @Override
    public void close() {
        rollback();
        session.close();
        client.close();
    }

    /**
     * Drop the database associated with this transaction. Calling this method will automatically rollback the
     * transaction.
     */
    protected void drop() {
        rollback();
        client.getDatabase(databaseName).drop();
    }
}
