package lc.esp.eo;

import lc.esp.eo.sdk.EO;
import lc.esp.eo.sdk.EOId;
import lc.mecha.log.MechaLogger;
import lc.mecha.log.MechaLoggerFactory;
import org.junit.jupiter.api.Test;

import java.io.ByteArrayOutputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Map;
import java.util.UUID;

/**
 * JUnit tests to exercise the EO system.
 *
 * @author Alex Leigh
 * @since mk20 (TOKYO PROTOCOL)
 */
public class EOTransactionTest {
    private static final MechaLogger logger = MechaLoggerFactory.getLogger(EOTransactionTest.class);

    @Test
    public void testObjCreate() {
        try (EOClient client = new EOClient()) {
            try (EOTransaction txn = client.getTransaction("junit_" + UUID.randomUUID(), null)) {
                try {
                    EOId id = txn.objCreate("test", Map.of("first", "Alex"));
                    EO obj = txn.objRead(id);
                    assert (obj != null);
                    txn.rollback();
                    obj = txn.objRead(id);
                    assert (obj == null);
                } finally {
                    txn.drop();
                }
            }
        }
    }

    @Test
    public void testFileCreate() throws IOException {
        try (EOClient client = new EOClient()) {
            try (EOTransaction txn = client.getTransaction("junit_" + UUID.randomUUID(), null)) {
                try {
                    // We basically just assume this will always be around in the build environment
                    try (FileInputStream fis = new FileInputStream("build.gradle")) {
                        String fileId = txn.fsFileCreate("test", fis, "text/plain");
                        logger.info("Created file: {}", fileId);
                        ByteArrayOutputStream bos = new ByteArrayOutputStream();
                        txn.fsFileRead(bos, "test");
                        bos.close();
                        logger.info("Read file: {}", bos.size());
                        assert (bos.size() > 0);
                        txn.rollback();
                    }
                } finally {
                    txn.drop();
                }
            }
        }
    }

    @Test
    public void testAssocCreate() throws IOException, InterruptedException {
        try (EOClient client = new EOClient()) {
            try (EOTransaction txn = client.getTransaction("junit_" + UUID.randomUUID(), null)) {
                try {
                    txn.initialize();
                    EOId alexId = txn.objCreate("person", Map.of("first", "Alex"));
                    EOId nikkiId = txn.objCreate("person", Map.of("first", "Nikki"));
                    txn.assocCreate(alexId, nikkiId, "boyfriend");
                    txn.assocCreate(alexId, nikkiId, "girlfriend");
                    txn.commit();
                } finally {
                    txn.drop();
                }
            }
        }
    }

    @Test
    public void testObjDelete() {
        try (EOClient client = new EOClient()) {
            try (EOTransaction txn = client.getTransaction("junit_" + UUID.randomUUID(), null)) {
                try {
                    EOId id = txn.objCreate("test", Map.of("first", "Alex"));
                    assert (id != null);
                    txn.objDelete(id);

                    EO obj = txn.objRead(id);
                    assert (obj == null);
                } finally {
                    txn.drop();
                }
            }
        }
    }
}